#!/usr/bin/python

'''
Python program written for Python 2.7 to analyse log files which have been 
created with an EZTap Pro RS-232 line analyser and its associated software.
When connected to a heater installation via a bus interface, the log files 
contain the complete traffic on the connected bus.
This program has been written to analyse the BS-Bus traffic log files and 
makes several assumptions based on knowledge of the BS-Bus protocol.  A 
different program logic is required to analyse the LP-Bus traffic.

Program features:
1. This program can analyse either the binary .dat files or ASCII .txt files
   which the EZView software produces.
2. Data bytes are either interpreted as-is or inverted. The RNEGATE variable
   controls the read-from-bus behavior, XNEGATE the send-to-bus behavior.
   Note: The dictionary addressing to decode source and destination device 
   address breaks when you use raw (already negated) telegram data with 
   RNEGATE = False.  Already negated data must be inverted to reflect the 
   real values.
3. To suppress certain telegrams (example: the Outside Temperature request), 
   define a SUPPRESS { } list of 4-digit numerical ProgNrs.
4. Dependent on your RS232 wiring, the EZTap sniffer may log the transmit AND
   the receive line.  Because the bus receiver immediately reads what is sent
   to the bus, you may get the characters echoed. Instead of modifying your 
   wiring, drop duplicate characters by setting ECHO = True.
5. PAUSEMIN is set to a somewhat longer duration than one character time. 
   This time determines when the program assumes that a telegram has ended and
   a new telegram begins.  Modify it only if the program does a bad job sepa-
   rating telegrams. A value of 3,000 is almost (!) always sufficient. 
   I have seen inter-character delays of 3,334 us within a telegram. The
   EZTap Pro analyser does a great job sniffing and time-stamping the serial
   bus communication.
6. Program output can be written to a file when the 'WRITE2FILE' variable is
   set to 'True'.  Program output to stdout is always enabled.
7. If you set UNITTEST = True, the program reads a log which you have created
   from listening to a real communication between the heater devices. It
   replays the telegrams on two serial PC ports.  One PC port connects to the
   serial interface of the BS-/LP-Bus interface board; the program sends
   GET and SET telegrams on this interface 'to the heater'. The other PC port
   writes the 'device responses' directly to the bus. It toggles a MOSFET ON 
   and OFF which pulls the bus low when it conducts. 
   This unit test setup CAN REPLAY BOTH SIDES OF THE BUS TRAFFIC, the requester
   and the responder without a real heater installation.
   To be clear: the 'bus' is not the real heater installation bus. It is a 
   wire with a pull-up resistor connected to +15V.  The BS-/LP-Bus interface 
   can read and write from and to this 'bus'. This setup helps me to test and 
   fine-tune my interface hardware and software on my lab bench and don't need
   to go to my heater installation which is kilometers away (think metric).
8. TPAUSE determines the pause between two telegrams in seconds. The minimum 
   length depends on whatever preprocessing the interface board performs and
   how much time this consumes.  If the interface receives a telegram in full
   before it begins to process it, the recommended duration is at least 32 
   character times (the max. telegram transport duration).
9. The global variables fout, serialPC and serialBus define byte streams and 
   must not be touched. 
   
   I put this program into the public domain. If you use it, you alone are 
   responsible to determine whether it is fit for the intenden purpose. I do
   not know for what this program will or could be used ad do not accept any 
   responsibility and liability for errors or any damage whatsoever which 
   results from its use.

   Like most software, this program can and probably does contain errors. I do 
   not claim that it will work under all circumstances in which you will use it.
'''

import sys
import os
import serial       # communication via serial interfaces
import time         # for sleep()
import math         # modulo arithmetic
import datetime     # Calculate date and time and differences thereof
import locale       #
from datetime import timedelta
from CRCstd   import crc16xmodem
from FieldIDs import FI, ADDRESS

# Some global variables
RNEGATE     = True  # invert received data bytes
XNEGATE     = True  # invert data bytes before sending
SUPPRESS    = { }   # 4-digit ProgNrs to suppress in the output
ECHO        = True  # each byte is echo'ed in the log file
PAUSEMIN    = 3400  # no characters received during this many microseconds
                    # 11/4800 + 15%  = one character time + 15%
WRITE2FILE  = True  # program output to file? True/False


# Replay the log file contents on two serial interfaces. Send all SET and GET
#   requests over a serial link to a serial port of the BS-Bus interface board.
#   This is the normal operation of a computer connected to the bus interface.
# Send all ACK, RET, and INF responses to a circuit which modulates the BS-Bus
#   or LP-Bus and therewith emulates a heating system response. 
# Both together is an exact replay of previously logged bus traffic.
#
# Note: This feature is not yet implemented.
#
UNITTEST    = True # replay the log file on two serial interfaces? True/False
TPAUSE      = 0.1

# ********* No user-serviceable parts below this point. *********


fout        = None  # output file stream. DO NOT TOUCH.
flog        = None  # log file stream. DO NOT TOUCH.
serialPC    = None  # possibly create a serial I/O stream later. DO NOT TOUCH.
serialBus   = None  # possibly create a serial I/O stream later. DO NOT TOUCH.



# -----------------------------------------------------------------------------
class PythonVersionError(Exception):
  ''' 
  Our own exeception if the program is not run under the correct Python version
  '''
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)

# -----------------------------------------------------------------------------
class CmdLineException(Exception):
  ''' Our own exeception for invalid command line arguments '''
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)
  
# -----------------------------------------------------------------------------
class InvalidAnalysisType(Exception):
  '''
  Our own exeception for invalid input data format.
  The format must be binary (.dat extension) or text (.txt extension).
  '''
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)
  
# -----------------------------------------------------------------------------
class PassParamError(Exception):
  '''
  Our own exeception for invalid pass parameter format.
  '''
  def __init__(self, value):
    self.value = value
  def __str__(self):
    return repr(self.value)
  

# -----------------------------------------------------------------------------
def recType ( binLine ):
  '''
  Translate a numerical record type to a human-readable string.
  '''
  rTypes     = { 2:"INF", 3:"SET", 4:"ACK", 5:"TBD", 6:"GET", 7:"RET", 8:"ERR" }
  tt = "unknown"
  
  if len(binLine) > 4:
    #print 'Type indicator is %i' % ord(binLine[4])
    if ord(binLine[4]) in rTypes:
      tt = rTypes[ord(binLine[4])]

  # Python 2.x converts only ASCII character 0..127 to bytes to write them to
  # a bytestream (output file).  We have to explicitly tell which encoding to
  # use.  
  return unicode(tt).encode('utf8')  # return the type string


# -----------------------------------------------------------------------------
def Dword2Long( instr4 ):
  '''
  Convert a four-byte string to a long integer
  '''
  if 4 != len(instr4):
    raise PassParamError('Dword2Long() expects a four-byte input.')

  if 'little' == sys.byteorder:
    return 2**24*ord(instr4[3])+2**16*ord(instr4[2])+2**8*ord(instr4[1])+ord(instr4[0])
  else:
    return 2**24*ord(instr4[0])+2**16*ord(instr4[1])+2**8*ord(instr4[2])+ord(instr4[3])
  
  
# -----------------------------------------------------------------------------
def Filetime2DateTime( instr8 ):
  '''
  Convert an eight-byte FILETIME struct to date and time. 
  FILETIME is a 64-bit Windows operating system structure:
     typedef struct _FILETIME {
       DWORD dwLowDateTime;
       DWORD dwHighDateTime;
     } FILETIME, *PFILETIME;
  A file time is a 64-bit value that represents the number of 100-nanosecond
  intervals that have elapsed since 12:00 A.M. January 1, 1601 Coordinated
  Universal Time (UTC)
  Remark: EZView timestamps each byte in this format.
  '''
  if 8 != len(instr8):
    raise PassParamError('Filetime2DateTime() expects an eight-byte input.')

  # Convert the low part to a long integer
  nLow  = long(Dword2Long(instr8[0:4]))
  
  # Convert the high part to a long integer and multiply by 2**32
  nHigh = long(Dword2Long(instr8[4:8]) << 32)
  
  # Build a 64-bit FILETIME value
  ft64  = nHigh + nLow
  
  # Calculate number of days since 1601-JAN-01
  d = ft64 / (86400 / 0.0000001)
  days    = math.modf(d)[1]
  fracday = math.modf(d)[0]
  #print 'days      = %i ' % days
  #print 'remainder = %f ' % fracday
  
  # Calculate number of seconds since the begin of this day
  s = fracday *86400
  seconds = math.modf(s)[1]
  microsec = (math.modf(s)[0] * 1000000)
  #print 'seconds   = %i ' % seconds
  #print 'microsec  = %i ' % microsec
  
  # Create a datetime object and return it
  BeginOfTime = datetime.datetime(1601,1,1,0,0,0,0)
  TimeDelta   = timedelta(days, seconds, microsec)
  FileTime    = BeginOfTime + TimeDelta

  return FileTime

# -----------------------------------------------------------------------------
def bin2hex( binstring ):
  '''
  Convert a sequence of bytes into their hex representation.
  No spaces are added between the hex bytes.
  Pass parameter:
    instring    a sequence of bytes
  Return value:
    string s   in hex
  '''
  s=''
  for x in range(0, len(binstring)):
    s = s + '%0.2x' % ord(binstring[x])
  return s
  
  
# -----------------------------------------------------------------------------
def printLine ( binString ):
  ''' 
  Function to create one line of hex data, separated by spaces 
  '''
  line = ''    # start with a clean slate
  
  # Fill the line with hex representation. Each data byte 
  # expands to a two-character hex tuple and a space.
  for c in range(0,len(binString)):
    line = line + '%0.2x' % ord(binString[c]) + ' '
  # endfor
    
  return line       
    

# -----------------------------------------------------------------------------
def schedule(fidi, payload):
  '''
  Disect the schedule payload, a string of bytes
  Pass parameters:
    int   fieldID
    string payload
  Returns:
    List of integers, comprised of six hours and minutes pairs
      if a correct FieldID has been found; an empty list otherwise.
  '''
  
  sched = False
  #print "Working in schedule() with fidi=%i and payload = %s." % (fidi, payload)
  if fidi == 0x053d0a8c or fidi == 0x3d050a8c:
    #print "Found 0x053d0a8c or 0x3d050a8c"
    sched = True
  if fidi == 0x053d0a8d or fidi == 0x3d050a8d:
    #print "Found 0x053d0a8d or 0x3d050a8d"
    sched = True
  if fidi == 0x053d0a8e or fidi == 0x053d0a8e:
    #print "Found 0x053d0a8e or 0x3d050a8e"
    sched = True
  if fidi == 0x053d0a8f or fidi == 0x3d050a8f:
    #print "Found 0x053d0a8f or 0x3d050a8f"
    sched = True
  if fidi == 0x053d0a90 or fidi == 0x3d050a90:
    #print "Found 0x053d0a90 or 0x3d050a90"
    sched = True
  if fidi == 0x053d0a91 or fidi == 0x3d050a91:
    #print "Found 0x053d0a91 or 0x3d050a91"
    sched = True
  if fidi == 0x053d0a92 or fidi == 0x3d050a92:
    #print "Found 0x053d0a92 or 0x3d050a92"
    sched = True

  if sched == True:
    #print "Found SCHEDULE fidi=0x%06x and payload=%s." % (fidi, payload)
    hr1  = int(payload[0:2], 16)
    if hr1 > 23:  # inactive, or invalid 
      hr1 = 0 
    min1 = int(payload[2:4], 16)
    hr2  = int(payload[4:6], 16)
    if hr2 > 23:  # inactive, or invalid 
      hr2 = 0 
    min2 = int(payload[6:8], 16)
    hr3  = int(payload[8:10], 16)
    if hr3 > 23:  # inactive, or invalid 
      hr3 = 0 
    min3 = int(payload[10:12], 16)
    hr4  = int(payload[12:14], 16)
    if hr4 > 23:  # inactive, or invalid 
      hr4 = 0 
    min4 = int(payload[14:16], 16)
    hr5  = int(payload[16:18], 16)
    if hr5 > 23:  # inactive, or invalid 
      hr5 = 0 
    min5 = int(payload[18:20], 16)
    hr6  = int(payload[20:22], 16)
    if hr6 > 23:  # inactive, or invalid 
      hr6 = 0 
    min6 = int(payload[22:24], 16)
    '''
    print 'P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i' % \
    (hr1, min1, hr2, min2, hr3, min3, hr4, min4, hr5, min5, hr6, min6)
    '''
    li = [ hr1, min1, hr2, min2, hr3, min3, hr4, min4, hr5, min5, hr6, min6 ]
  else:
    li = []  # regognised no valid FieldID
    
  return li  # return a list of three begin-end pairs


# -----------------------------------------------------------------------------
def dateandtime(fidi, payload):
  
  foundDateTime = False
  #print "Working in datenadtime() with fidi=%i and payload = %s." % (fidi, payload)
  if fidi == 0x0500006c:
    foundDateTime = True

  if foundDateTime == True:
    #print "Found DATETIME fidi=0x%.06x and payload=%s." % (fidi, payload)
    yr  = int(payload[2:4], 16) + 1900
    mo  = int(payload[4:6], 16)
    dd  = int(payload[6:8], 16)
    
    flag= int(payload[8:10], 16)

    hr   = int(payload[10:12], 16)
    min  = int(payload[12:14], 16)
    sec  = int(payload[14:16], 16)
    unk  = int(payload[16:18], 16)
    '''
    print 'Date: %.04i-%.02i-%.02i Flag:0x%.02x Time: %.02i:%.02i:%.02i  unknown:%.02x' % \
    (yr, mo, dd, flag, hr, min, sec, unk)
    '''

    li = [ yr, mo, dd, hr, min, sec ]
  else:
    li = []  # return an empty list
    
  return li  # return date and time integers



# =============================================================================
#                        processTelegram()
# =============================================================================
def processTelegram( fo, ezLine, offset, pause):
  
  # Debug info can be useful if you get telegram separation problems.
  #print 'DEBUG ezLine = %s, len=%i' % (printLine(ezLine), len(ezLine))
  
  # Plausibility check: minimum length
  if len(ezLine) > 10:    # plausibility check
    o = printLine(ezLine[0:]) #
    negate = ''
    for i in range (0,len(ezLine)):
      byte = ord(ezLine[i])^255
      negate += chr(byte)
    o_neg = printLine(negate)    
    us   = int(offset)
    sec  = int(offset)/1000000
    rus  = int(offset) % 1000000
    hours= sec/3600
    rsec = sec % 3600
    min  = rsec/60
    sec  = rsec % 60

    sched   = []                   # declaration for later use
    doPrint = True                 # declaration for later use
    tt      = recType( ezLine )    # record type
    fid     = bin2hex(ezLine[5:9]) # fieldID, hex format
    fidi = int(fid,16)
    fidr = int(fid[2:4]+fid[0:2]+fid[4:],16)
    #print   'DEBUG fid=%s fidi=%.08x, fidr=%.08x\n' % (fid, fidi, fidr)
    #f.write('DEBUG fid=%s fidi=%.08x, fidr=%.08x\n' % (fid, fidi, fidr))
    payload = bin2hex(ezLine[9:-2])
    
    
    # Attempt to extract date and time from the payload
    dt = []
    dt   = dateandtime(fidi, payload)
    
    if fidi in FI:
      pn  = format("%i" % FI[fidi][0])  # ProgNr
    elif fidr in FI:
      pn  = format("%i" % FI[fidr][0])  # ProgNr
    else:
      pn = 'UNKNOWN_PN'
      
    # Do we want to suppress the printout for this ProgNr?
    if pn not in SUPPRESS:
      print   'Pause of '+'{0:,}'.format(pause)+' us. '
      print   'Offset (hh:mm:ss.us) is now: %0.2i:%0.2i:%0.2i.%i'   % (hours, min, sec, rus)
      print   'INV %s len=0x%.02x' % ( o, len(ezLine))
      print   'RAW %s len=0x%.02x' % ( o_neg, len(negate))
      if WRITE2FILE:
        fo.write('Pause of '+'{0:,}'.format(pause)+' us. ')
        fo.write('Offset (hh:mm:ss.us) is now: %0.2i:%0.2i:%0.2i.%i\n' % (hours, min, sec, rus))
        fo.write( 'INV %s len=0x%.02x\n' % ( o, len(ezLine)) )
        fo.write( 'RAW %s len=0x%.02x\n' % ( o_neg, len(negate) ))
    
      # Make a series of decisions to determine the telegram type
      if tt == 'INF':
        #f.write('DEBUG fid=%s fidi=0x%.08x, fidr=0x%.08x\n' % (fid, fidi, fidr))
        if fidi in FI: # byte order like in command table
          print   'fieldID (INF) = 0x%.08x'   % fidi          
          print    "Data point is %s (ProgNr %s)."   % (FI[fidi][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (INF) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidi][1].encode('utf_8'), pn ) )
              
        # Try the same with the first two bytes reversed
        elif fidr in FI:  # swapped byte order
          print   'fieldID (WARNING reverse the first two bytes in the command reference!) (INF) = 0x%.08x'   % fidr
          print    "Data point is %s (ProgNr %s)."   % (FI[fidr][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (WARNING reverse the first two bytes in the command reference!) (INF) = 0x%.08x\n' % fidr)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidr][1].encode('utf_8'), pn ) )
                
        else:
          print   'fieldID (INF) = 0x%.08x (ProgNr %s)'   % (fidi,pn)
          if WRITE2FILE:
            fo.write('fieldID (INF) = 0x%.08x (ProgNr %s)\n' % (fidi,pn))

        # Is this a date-and-time telegram?
        if (len(dt) == 6):
          print   'Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i' % \
          (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
          if WRITE2FILE:
            fo.write('Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i\n' % \
            (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5]))

        # Send binary data to the bus for unit tests
        if UNITTEST:
          # We could insert real-life time delays here because we know the 
          # pause duration in us. Syntax: time.sleep(pause/1000000)
          if XNEGATE:
            serialBus.write(negate)
            print      'Unittest XNEGATE, Bus sent INF <<< %s'   % o_neg
            flog.write('Unittest XNEGATE, Bus sent INF <<< %s\n' % o_neg)
          else:
            serialBus.write(ezLine)    # send binary data
            print      'Unittest,         Bus sent INF <<< %s'   % o
            flog.write('Unittest,         Bus sent INF <<< %s\n' % o)


      elif tt == 'ACK':
        if fidi in FI:    # byte order like in command table
          print   'fieldID (ACK) = 0x%.08x'   % fidi
          print    "Data point is %s (ProgNr %s)."   % (FI[fidi][1], pn)
          if write2File:
            fo.write('fieldID (ACK) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidi][1].encode('utf_8'), pn ) )

        # Try the same with the first two bytes reversed
        elif fidr in FI:  # swapped byte order
          print   'fieldID (WARNING reverse the first two bytes in the command reference!) (ACK) = 0x%.08x' % fidr
          print    "Data point is %s (ProgNr %s)." % (FI[fidr][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (WARNING reverse the first two bytes in the command reference!) (ACK) = 0x%.08x\n' % fidr)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidr][1].encode('utf_8'), pn ) )
                
        else:
          print   'fieldID (ACK) = 0x%.08x (ProgNr %s)'   % (fidi,pn)
          if WRITE2FILE:
            fo.write('fieldID (ACK) = 0x%.08x (ProgNr %s)\n' % (fidi,pn))

        # Send binary data to the bus for unit tests
        if UNITTEST:
          # We could insert real-life time delays here because we know the 
          # pause duration in us. Syntax: time.sleep(pause/1000000)
          if XNEGATE:
            serialBus.write(negate)
            print      'Unittest XNEGATE, Bus sent ACK <<< %s'   % o_neg
            flog.write('Unittest XNEGATE, Bus sent ACK <<< %s\n' % o_neg)
          else:
            serialBus.write(ezLine)    # send binary data
            print      'Unittest,         Bus sent ACK <<< %s'   % o
            flog.write('Unittest,         Bus sent ACK <<< %s\n' % o)

      # The fieldID is defined for a RET type telegram instead of 
      # a GET type telegram, where the first two bytes are swapped.
      elif tt == 'GET':   # GET type telegram, bytes are swapped
        #f.write('DEBUG fid=%s fidi=0x%.08x, fidr=0x%.08x\n' % (fid, fidi, fidr))
        if fidi in FI:    # byte order like in command table
          sched = schedule(fidi, payload)    # try to decode
          print   'fieldID (WARNING reverse the first two bytes in the command reference!) (GET) = 0x%.08x'   % fidi
          print    "Data point is %s (ProgNr %s)." % (FI[fidi][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (WARNING reverse the first two bytes in the command reference!) (GET) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidi][1].encode('utf_8'), pn ) )
                
        # Try the same with the first two bytes reversed
        elif fidr in FI:  # swapped byte order
          sched = schedule(fidi, payload)    # try to decode a schedule
          print   'fieldID_rev (GET) = 0x%.08x'   % fidr
          print    "Data point is %s (ProgNr %s)." % (FI[fidr][1], pn)
          if WRITE2FILE:
            fo.write('fieldID_rev (GET) = 0x%.08x\n' % fidr)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidr][1].encode('utf_8'), pn ) )
                
        else:
          print   'fieldID (GET) = 0x%.08x (ProgNr %s)'   % (fidi,pn)
          if WRITE2FILE:
            fo.write('fieldID (GET) = 0x%.08x (ProgNr %s)\n' % (fidi,pn))

        # Is this a schedule telegram?
        if (len(sched) == 12):
          print   'P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i' % \
          (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11])
          if WRITE2FILE:
            fo.write('P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i\n' % \
            (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11]))

        # Is this a date-and-time telegram?
        if (len(dt) == 6):
          print     'Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i'   % (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
          if WRITE2FILE:
            fo.write('Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i\n' % (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5]))
          
        # Send binary data to the DUT for unit tests
        if UNITTEST:
          # We could insert real-life time delays here because we know the 
          # pause duration in us. Syntax: time.sleep(pause/1000000)
          if XNEGATE:
            serialPC.write(negate)
            print      'Unittest XNEGATE, PC  sent GET >>> %s'   % o_neg
            flog.write('Unittest XNEGATE, PC  sent GET >>> %s\n' % o_neg)
          else:
            serialPC.write(ezLine)    # send binary data
            print      'Unittest,         PC  sent GET >>> %s'   % o
            flog.write('Unittest,         PC  sent GET >>> %s\n' % o)

      elif tt == 'RET':  # is a response type telegram
        # The response telegram reverses the fieldID byte order of the 
        # first two bytes.
        # The bsbgateway software seems to define the fieldID for the response
        # type (RET) telegram, not for the GET telegram.
        #f.write('DEBUG fid=%s fidi=0x%.08x, fidr=0x%.08x\n' % (fid, fidi, fidr))
        if fidi in FI:    # byte order like in command table
          sched = schedule(fidi, payload)  # try to decode
          print    'fieldID (RET) = 0x%.08x' % fidi
          print    "Data point is %s (ProgNr %s)." % (FI[fidi][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (RET) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidi][1].encode('utf_8'), pn ) )
 
        # Reverse the first two bytes of the fieldID
        elif fidr in FI:  # swapped byte order
          sched = schedule(fidi, payload)  # try to decode
          print   'fieldID (WARNING reverse the first two bytes in the command reference!) (RET) = 0x%.08x' % fidi
          print    "Data point is %s (ProgNr %s)." % (FI[fidr][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (WARNING reverse the first two bytes in the command reference!) (RET) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidr][1].encode('utf_8'), pn ) )
      
        else:
          print   'fieldID (RET) = 0x%.08x (ProgNr %s)'   % (fidi,pn)
          fo.write('fieldID (RET) = 0x%.08x (ProgNr %s)\n' % (fidi,pn))
      
        if (len(sched) == 12):
          print   'P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i' % \
          (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11])
          if WRITE2FILE:
            fo.write('P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i\n' % \
            (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11]))

        # Is this a date-and-time telegram?
        if (len(dt) == 6):
          print   'Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i' % \
          (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
          if WRITE2FILE:
            fo.write('Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i\n' % \
            (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5]))

        # Send binary data to the bus for unit tests
        if UNITTEST:
          # We could insert real-life time delays here because we know the 
          # pause duration in us. Syntax: time.sleep(pause/1000000)
          if XNEGATE:
            serialPC.write(negate)
            print      'Unittest XNEGATE, Bus sent RET <<< %s'   % o_neg
            flog.write('Unittest XNEGATE, Bus sent RET <<< %s\n' % o_neg)
          else:
            serialPC.write(ezLine)    # send binary data
            print      'Unittest,         Bus sent RET <<< %s'   % o
            flog.write('Unittest,         Bus sent RET <<< %s\n' % o)

      elif tt == 'SET':   # bytes are swapped
        if fidi in FI:    # byte order like in command table
          sched = schedule(fidi, payload)    # try to decode a schedule
          print   'fieldID (WARNING reverse the first two bytes in the command reference!) (SET) = 0x%.08x' % fidi
          print    "Data point is %s (ProgNr %s)." % (FI[fidi][1], pn)
          if WRITE2FILE:
            fo.write('fieldID (WARNING reverse the first two bytes in the command reference!) (SET) = 0x%.08x\n' % fidi)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidi][1].encode('utf_8'), pn ) )

        # Reverse the first two bytes of the fieldID
        elif fidr in FI:  # swapped byte order
          sched = schedule(fidi, payload)    # try to decode a schedule
          print   'fieldID_rev (SET) = 0x%.08x' % fidr
          print    "Data point is %s (ProgNr %s)." % (FI[fidr][1], pn)
          if WRITE2FILE:
            fo.write('fieldID_rev (SET) = 0x%.08x\n' % fidr)
            fo.write( "Data point is %s (ProgNr %s).\n" % (FI[fidr][1].encode('utf_8'), pn ) )
                
        else:
          print   'fieldID (SET) = 0x%.08x (ProgNr %s)'   % (fidi,pn)
          if WRITE2FILE:
            fo.write('fieldID (SET) = 0x%.08x (ProgNr %s)\n' % (fidi,pn))

        if len(sched) == 12:
          print   'P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i' % \
          (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11])
          if WRITE2FILE:
            fo.write('P1: %.02i:%.02i-%.02i:%.02i; P2: %.02i:%.02i-%.02i:%.02i; P3: %.02i:%.02i-%.02i:%.02i\n' % \
            (sched[0], sched[1], sched[2], sched[3], sched[4], sched[5], sched[6], sched[7], sched[8], sched[9], sched[10], sched[11]))

        # Is this a date-and-time telegram?
        if len(dt) == 6:
          print   'Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i' % \
          (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5])
          if WRITE2FILE:
            fo.write('Date: %.04i-%.02i-%.02i  Time: %.02i:%.02i:%.02i\n' % \
            (dt[0], dt[1], dt[2], dt[3], dt[4], dt[5]))
            
        # Send binary data to the DUT for unit tests
        if UNITTEST:
          # We could insert real-life time delays here because we know the 
          # pause duration in us. Syntax: time.sleep(pause/1000000)
          if XNEGATE:
            serialPC.write(negate)
            print      'Unittest XNEGATE, PC  sent SET >>> %s'   % o_neg
            flog.write('Unittest XNEGATE, PC  sent SET >>> %s\n' % o_neg)
          else:
            serialPC.write(ezLine)    # send binary data
            print      'Unittest,         PC  sent SET >>> %s'   % o
            flog.write('Unittest,         PC  sent SET >>> %s\n' % o)

      # endif, decisions to determine the telegram type

      # Extract the CRC from the telegram
      crc = '0x' + bin2hex(ezLine[-2:])  # sniffed, is a string
          
      # Calculate the CRC and compare. The crc16modem() function returns
      # a number which needs to be converted to hex and zero-padded.
      crc2 = "0x{0:04x}".format(crc16xmodem(ezLine[:-2])) # calculated CRC, i s a number
      if crc != crc2:
        # *** CRC ERROR ***
        print   'CCITT-XMODEM CRC is %s' % crc2
        if WRITE2FILE:
          fo.write('CCITT-XMODEM CRC is %s\n' % crc2)

      # Print some info about the telegram. Texts may contain non-ASCII characters!
      src = ADDRESS[ord(ezLine[1])&0x7F]
      #print 'src is ', printLine(src)
      dst = ADDRESS[ord(ezLine[2])]
      #print 'dst is ', printLine(dst)
      print   'Src=%0.2x/%s Dest=%0.2x/%s Type=%s FieldID=%s Payload=%s CRC=%s\n'   % ( ord(ezLine[1])&0x7F, src, ord(ezLine[2]), dst, tt, fid, payload, crc )
      if WRITE2FILE:
        fo.write('Src=%0.2x/%s Dest=%0.2x/%s Type=%s FieldID=%s Payload=%s CRC=%s\n\n' % ( ord(ezLine[1])&0x7F, unicode(src).encode('utf8'), ord(ezLine[2]), unicode(dst).encode('utf8'), tt, fid, payload, crc ) )
  
    else:
      print   '.'       # instead of suppressed telegram statistics
      if WRITE2FILE:
        fo.write('.\n')    # instead of suppressed telegram statistics
    
  # endif, len(ezLine > 10)
  else:
    raise PassParamError(len(ezLine))
  
  # Instead of real-life delays, insert a constant but sufficiently long delay.
  # Recommended duration: minimum 32 character times (= max. telegram duration)
  if UNITTEST:
    time.sleep(TPAUSE)    # pause between telegrams
    
# --- processTelegram() ---  


  
        
      
# =============================================================================
#                               main()
# =============================================================================
def main( argv ):

  # Some common variables
  isDat      = 0
  isTxt      = 0
  pauseMin   = 0
  pauseMax   = 0
  pauseAvg   = 0
  ezTs       = 0    # holds the timestamp of the first transmitted byte
  telegrams  = 0    # how many telegrams have been processed?
  inFile     = ''   # input filename
  
  global fout       # output file descriptor, processed EZView output
  global flog       # log file descriptor
  global serialPC   # serial port descriptor, PC to DUT
  global serialBus  # serial port descriptor, PC to bus
  
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  try:
    
    PY2  = sys.version_info[0] == 2
    PY3  = sys.version_info[0] == 3
    if PY2 != True:
      raise PythonVersionError('This program must be run under Python version 2.x')
  
    if (len(sys.argv) != 2):
      raise CmdLineException('You need to supply a filename.')

    inFile = sys.argv[1]
    
    # Did the user supply a filename with a .dat extension?
    if 'dat' == inFile.split('.')[-1]:
      fin     = open(inFile,'rb')    # open EZView log file for binary read
      outFile = inFile.split('.')[:-1][0]+'-ana_dat.txt'
      isDat = 1
      
    # Did the user supply a filename with a .txt extension?
    if 'txt' == inFile.split('.')[-1]:
      fin     = open(inFile,'r')    # open EZView log file for read
      isTxt   = 1
      outFile = inFile.split('.')[:-1][0]+'-ana_txt.txt'
     
    if ((isDat + isTxt) == 0):
      raise CmdLineException("The file extension must be .dat (binary file) or .txt (ASCII file)")
    
    if WRITE2FILE:
      fout = open(outFile,'w')    # where we write the formatted output
    else:
      fout = None
    if fout == None:
      print '******* ATTENTION: Not writing output to file. *******'
    else:
      print '******* ATTENTION: Writing output to file. *******'
      
    print '----------------------------------------------------'
    print 'inFile  is: ' + inFile
    print 'outFile is: ' + outFile 
    if WRITE2FILE:
      fout.write('----------------------------------------------------\n')
      fout.write('inFile  is: %s\n' % inFile)
      fout.write('outFile is: %s\n' % outFile )
    
    if UNITTEST == True:
      serialPC  = serial.Serial('/dev/ttyUSB0', 4800, parity=serial.PARITY_NONE)  # open serial port #1
      serialBus = serial.Serial('/dev/ttyS0', 4800, parity=serial.PARITY_ODD)   # open serial port #2
      logFile   = './UT-log.txt'
      flog = open(logFile, 'w')   # open a log file
      flog.write('Processing input log file %s\n' % inFile)
      flog.write('UNITTEST: Sending unit test data to the BS-/LP-Bus interface via %s\n' % serialPC.name)
      flog.write('UNITTEST: Sending unit test data to the BS-/LP-Bus wire      via %s\n' % serialBus.name)
      if XNEGATE:
        flog.write('UNITTEST: Inverting log file data before sending them.\n\n')
      print '\nUNITTEST: Sending unit test data to the BS-/LP-Bus interface via %s' % serialPC.name
      print 'UNITTEST: Sending unit test data to the BS-/LP-Bus wire      via %s\n' % serialBus.name
      if WRITE2FILE:
        fout.write('\nUNITTEST: Sending unit test data to the BS-/LP-Bus interface via %s\n' % serialPC.name)
        fout.write('UNITTEST: Sending unit test data to the BS-/LP-Bus wire      via %s\n\n' % serialBus.name)

    # *******************************************************************
    # Process EZView binary log files and interpret them as BS-Bus traces
    # *******************************************************************
    if isDat == 1:
      # Start at file buffer offset 68. Loop over ezNumRecords in the file,
      # each record is ezRecLen long
      index    = 68     # begin of EZ-Pro records in file
      ezSender = 0      # direction is still undetermined
      numBytes = 0      # bytes in this transmission
      ezTs     = 0      # timestamp of first byte, start time
      ezTi     = 0      # this byte timestamp
      ezTp     = 0      # previous byte timestamp
      ezLine   = ''     # output line
      
      # Read the whole file in one swoop
      ezDataB = fin.read()    # read as one byte string from input file
      
      # All the DWORDS in a log file are little-endian. Convert them to decimal
      ezHdrLen     = Dword2Long(ezDataB[32:36])
      ezVersion    = Dword2Long(ezDataB[36:40])
      ezDeviceType = Dword2Long(ezDataB[40:44])
      ezNumRecords = Dword2Long(ezDataB[44:48])
      ezRecLen     = Dword2Long(ezDataB[48:52])
      # FILETIME is a structure; it contains a 64-bit value which signifies
      # the number of 100ns intervals since 1601-JAN-01.  Convert the 64-bit
      # elapsed time to an absolute date and time.
      ezCollStart  = Filetime2DateTime(ezDataB[52:60])
      ezCStart     = ezCollStart.strftime("%Y-%m-%d %H:%M:%S.%f")
      ezDataStart  = Filetime2DateTime(ezDataB[60:68])
      ezDStart     = ezDataStart.strftime("%Y-%m-%d %H:%M:%S.%f")
      
      # Display some settings
      #print 'The EZView file header is %i bytes long.' % ezHdrLen
      #print 'EZView version is %i.' % ezVersion
      #print 'EZView device type is %i.' % ezDeviceType
      print 'The input file contains %i data and RS232 control signal records.' % (ezNumRecords)
      print 'Disregard all records with RS232 control signals.'
      print 'Disect the byte stream if a pause is longer than %i us (%.2f character time).' % (PAUSEMIN, PAUSEMIN / 2292.0)
      if RNEGATE == True:
        print '!!! Data are shown inverted !!!'
      else:
        print 'Data are not inverted.'
      #print 'Record length is %i.' % ezRecLen
            
      print '-------------------------------------------------------'
      print      'Data collection start time: %s'   % ezCStart
      print      'First data record time:     %s'   % ezDStart
      print '-------------------------------------------------------\n'
    
      if WRITE2FILE:
        #fout.write('The EZView file header is %i bytes long.\n' % ezHdrLen)
        #fout.write('EZView version is %i.\n' % ezVersion)
        #fout.write('EZView device type is %i.\n' % ezDeviceType)
        fout.write('The input file contains %i data and RS232 control signal records.\n' % ezNumRecords)
        fout.write('Disregard all records with RS232 control signals.\n')
        fout.write('Disect the byte stream if a pause is longer than %i us (%.2f character time).\n' % (PAUSEMIN,PAUSEMIN / 2292.0))
        if RNEGATE == True:
          fout.write('!!! Data are shown inverted !!!\n')
        else:
          fout.write('Data are not inverted.\n')
        #fout.write('Record length is %i.\n' % ezRecLen)
        fout.write('-------------------------------------------------------\n')        
        fout.write('Data collection start time: %s\n' % ezCStart )
        fout.write('First data record time:     %s\n' % ezDStart)
        fout.write('-------------------------------------------------------\n\n')
        
      # Each EZView log file record contains one data byte and various 
      # associated housekeeping information. 
      for count in range(0, ezNumRecords): # do for all EZView log file records
        
        # Read one EZ-Pro data record into the work buffer
        ezWork = ezDataB[index:index+ezRecLen]

        # I decided to print out only data records and disregard 
        # those records which indicate control signal changes.
        # Byte4=1 indicates a DTE data byte, Byte4=2 indicates a DCE data byte
        # Is this a data byte record?
        if 1 == ord(ezWork[4]) or 2 == ord(ezWork[4]):
          
          # Is a data record. Get the four-byte timestamp, convert it to decimal
          ezTi = Dword2Long(ezWork[0:4])  # timestamp for this data byte
        
          if count == 0:      # this is the first record
            ezTs   = ezTi     # start time against which we calculate offsets
            ezTp   = ezTi     # previous timestamp
            print 'Null reference timestamp ezTs is %d' % ezTs
            print '\nTelegram #%i begins with log file record %i:' % ((telegrams+1), (count+1))
            if WRITE2FILE:
              fout.write('Null reference timestamp ezTs is %d\n' % ezTs)
              fout.write('\nTelegram #%i begins with log file record %i:\n' % ((telegrams+1), (count+1)))
              

          # Calculate pause duration
          pause = ezTi - ezTp
          
          # Statistics: Keep track of maximum and minimum pause durations
          if pauseMin == 0:
            pauseMin = ezTi-ezTp
          if pauseMin > pause:
            pauseMin = pause    # remember as new minimum
          if pauseMax < pause:
            pauseMax = pause    # remember as new maximum
          pauseAvg += pause
                  
          # The current timestamp becomes the previous time 
          if ezTi >  ezTp:
            ezTp = ezTi         # this becomes now the previous timestamp
        
          # Get the DTE / DCE sender indicator
          thisDTEDCE = ord(ezWork[4])  # fifth byte
        
          # The first transmission initialises the DTE / DCE flag
          if ezSender == 0:         # sender is still undetermined
            ezSender = thisDTEDCE   # remember who started the communication
          # endif, the first transmitted byte. Now we know who started the communication.
        
          # Is the pause between two bytes longer than a given limit? 
          # I.e., has a new transmission block been detected?
          if pause > PAUSEMIN:      # no bytes received during this time
            # The pause in transmission is longer than one character time
            # First thing to do: print the previous transmission block.
            #print 'New transmission begins after %i us\n' % (pause)
            #fout.write('New transmission begins after %i us\n' % (pause))

            #o        = printLine(ezLine)
            offset   = locale.format("%d", ezTi - ezTs, grouping=True)
            
            processTelegram(fout, ezLine, offset, pause)
            
            # Start a new telegram
            telegrams += 1            
            print 'Telegram #%i begins with log file record %i:' % ((telegrams+1), (count+1))
            if WRITE2FILE:
              fout.write('Telegram #%i begins with log file record %i:\n' % ((telegrams+1), (count+1)))          
            ezLine   = ''
            numBytes = 0

          # Get the data byte from this record, collect them in ezLine
          ezByte = ord(ezWork[5])   # get data byte
          if RNEGATE == True:        # need to invert the byte?
            ezByte = ezByte^255     # yes, invert
          if ECHO != True:
            ezLine += chr(ezByte)   # append this data byte
          else:
            if numBytes % 2 == 0:
              ezLine += chr(ezByte) # append every other data byte
            numBytes = numBytes + 1 # count the data bytes in a transmission          
          
          numBytes = numBytes + 1   # count the data bytes in a transmission
          #print 'DEBUG Byte %i=%.02x numBytes=%i, ezTi-ezTp=%i\n' % (count,ezByte,numBytes,(ezTi-ezTp) )
          #fout.write('DEBUG Byte %i=%.02x numBytes=%i, ezTi-ezTp=%i\n' % (count,ezByte,numBytes,(ezTi-ezTp) ) )
          #print ezTi - ezTp        # DEBUG Show inter-character times
          ezTp = ezTi               # this becomes now the previous timestamp
                  
        # endif, this record was a data record

        # Increment the file buffer index for all processed records
        index = index + ezRecLen    # go to next data record in file buffer
        
      # endfor, process all input file records

      # Process the last remaining telegram which is not followed by a pause
      telegrams += 1
      print 'Telegram #%i begins with log record %i:' % (telegrams+1, (count+1))
      if WRITE2FILE:
        fout.write('Telegram #%i begins with log record %i:\n' % ((telegrams+1), (count+1)))
        
      processTelegram(fout, ezLine, offset, pause)

      # - - - - - - - - - -  end of binary input file handling  - - - - - - - -

      
      
    # *******************************************************************
    # Process EZView ASCII log files and interpret them as BS-Bus traces
    # *******************************************************************
    elif isTxt == 1:
      ezDataT = fin.read()      # read as one string from input file
      
      # Separate the lines in the input file. 
      ezData      = ezDataT.split('\n')
      seq         = []                 # declaration, will be used later
      ezNumBytes  = len(ezData)-1      # number of EZView data records in the file
      ezTs        = 0
      ezTi        = 0
      ezTp        = 0
    
      numBytes    = 0                  # number of bytes in one telegram
      ezLine      = ''
   
      print 'The input file contains %i data and RS232 control signal records.' % (ezNumBytes)
      print 'Disect the byte stream if a pause is longer than %i us (%.2f character time).' % (PAUSEMIN, PAUSEMIN / 2292.0)
      if WRITE2FILE:
        fout.write('The input file contains %i data and RS232 control signal records.\n' % (ezNumBytes))
        fout.write('Disect the byte stream if a pause is longer than %i us (%.2f character time).\n' % (PAUSEMIN,PAUSEMIN / 2292.0))
        
      
      for count in range(0, ezNumBytes):  # do for all EZView log file records
      
        # Get one input line and split it into fields
        # Each line in the log file consists of fields separated by TAB characters:
        # Fields are: LineNo, timestamp, None, None, hexByte, ASCII
        # Example: 
        # 2	04/25/16 15:35:01.514.708			75 	u
        linex   = ezData[count]         # one input line
        line    = linex.split('\t') # separate the fields
        hour    = line[1][9:11]
        #fout.write('hour = %s ' % hour)
        min     = line[1][12:14]
        #fout.write('min = %s ' % min)
        sec     = line[1][15:17]
        #fout.write('sec = %s ' % sec)
        us      = line[1][18:26].replace('.', '') # decimal number
        usT     = (int(hour)*3600 + int(min)*60 + int(sec)) * 1000000 + int(us)
      
        ezTi    = usT               # timestamp for this data byte
      
        if count == 0:              # first byte from log file
          ezTs = ezTi               # first timestamp
          ezTp = ezTi               # previous timestamp
          print 'Null reference timestamp ezTs is %d' % ezTs
          print '\nTelegram #%i begins with log file record %i:' % ((telegrams+1), (count+1))
          if WRITE2FILE:
            fout.write('Null reference timestamp ezTs is %d\n' % ezTs)
            fout.write('\nTelegram #%i begins with log file record %i:\n' % ((telegrams+1), (count+1)))
          
        # Calculate pause duration
        pause = ezTi - ezTp
      
        # Statistics: Keep track of the shortest pause time
        if pauseMin == 0:
          pauseMin = pause
        if pauseMin > pause:
          pauseMin = pause        # remember as new minimum
        if pauseMax < pause:
          pauseMax = pause        # remember as new maximum
        pauseAvg += pause

        if ezTi > ezTp:
          ezTp = ezTi             # this becomes the previous timestamp
      
        # Has a new telegram started?
        if pause > PAUSEMIN:
          #o        = printLine(ezLine)
          offset   = locale.format("%d", ezTi - ezTs, grouping=True)

          processTelegram(fout, ezLine, offset, pause)
          
          # Start a new telegram
          telegrams += 1
          print      'Telegram #%i begins with log file record %i:'   % ((telegrams+1), (count+1))
          if WRITE2FILE:
            fout.write('Telegram #%i begins with log file record %i:\n' % ((telegrams+1), (count+1)))
          ezLine   = ''             # empty ezLine
          numBytes = 0              # there are no bytes in ezLine
        
        # The fifth field is the data byte. Negate if requested.
        binByte = int(line[4],16)   # convert one data byte to binary
        if RNEGATE == True:
          binByte = binByte ^ 0xff  # negate the byte
        ezLine   += chr(binByte)
        numBytes += 1               # number of bytes in this telegram
        offset   = ezTi - ezTs

      # endfor, do for all input lines
    
      # Process the last remaining telegram which is not followed by a pause
      telegrams += 1
      print 'Telegram #%i begins with log record %i:' % ((telegrams+1), (count+1))
      if WRITE2FILE:
        fout.write('Telegram #%i begins with log record %i:\n' % ((telegrams+1), (count+1)))
                   
      processTelegram(fout, ezLine, offset, pause)
      
      
    # end of ASCII input file handling
    else:   # we should never come here
      raise InvalidAnalysisType("Unknown input format selected. Permissible are binary (.dat) and text (.txt)")
    
    # endelse, .dat or .txt input file
    
    pauseAvg = pauseAvg / count
    print '\n\nMinimum observed pause between telegrams was %i us = %f character times.' % (pauseMin, pauseMin/2291.666)
    print 'Maximum observed pause between telegrams was %i us = %f character times.' % (pauseMax, pauseMax/2291.666)
    #print 'Average observed pause between telegrams was %i us = %f character times.' % (pauseAvg, pauseAvg/2291.666)
    if WRITE2FILE:
      fout.write('\n\nMinimum observed pause between telegrams was %i us = %.2f character times.\n' % (pauseMin, pauseMin/2291.666))
      fout.write('Maximum observed pause between telegrams was %i us = %f character times.\n' % (pauseMax, pauseMax/2291.666))
      #fout.write('Average observed pause between telegrams was %i us = %f character times.\n' % (pauseAvg, pauseAvg/2291.666))
      
          
  # - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -
  except CmdLineException as e:
    print 'ERROR: ' , e.value
  
  except PythonVersionError as e:
    print 'ERROR: ' , e.value
   
  except:
    print "ERROR: ", sys.exc_info() [0]
    raise

  
  
  if os.access(inFile, os.R_OK):
    print 'Closing  input file %s' % inFile
    fin.close()
  if WRITE2FILE:
    print 'Closing output file %s' % outFile
    fout.close()
  if os.access(logFile, os.W_OK):
    print 'Closing    log file %s' % logFile
    flog.close()
  if UNITTEST:
    print 'Closing serial interfaces %s and %s' % (serialPC.name, serialBus.name)
    serialPC.close()
    serialBus.close()
  print 'Done.'
    
# -----------------------------------------------------------------------------

if __name__ == "__main__":
   main(sys.argv[1:])
